package com.jh.core.security.api;

import java.util.ArrayList;
import java.util.Base64;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

import javax.servlet.http.HttpSession;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.annotation.AuthenticationPrincipal;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.util.StringUtils;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.server.ResponseStatusException;

import com.jh.core.error.ApiException;
import com.jh.core.error.ErrorMessage;
import com.jh.core.security.AppUserAuthentication;
import com.jh.core.security.AuthError;
import com.jh.core.security.AuthService;
import com.jh.core.security.PasswordUtils;
import com.jh.core.security.model.AppUser;
import com.jh.core.utils.Utils;


@RestController
public class AuthApiController {
    
    /** The auth service. */
    @Autowired
    private AuthService authService;
    
    @Value("${auth.controller.signup.enabled:true}")
    private boolean signupEnabled;
    
    @Value("${auth.controller.reset.enabled:true}")
    private boolean resetEnabled;
    
    @RequestMapping({"/auth/logout"})
    @ResponseStatus(HttpStatus.NO_CONTENT)
    public void logoutApi(HttpSession session) {
		SecurityContextHolder.clearContext();
        if (session != null) {
            session.invalidate();
        }
    }

    @PostMapping(path = "/auth/login")
    public LoginResponse login(@RequestBody LoginRequest login) {
    	Utils.sleep(250);
    	Authentication auth = null;
    	try {
    		auth = authService.authenticate(login.getUsername(), login.getPassword());
    	} catch (AuthenticationException e) {
    		//do nothing - will be resolved later
    	}
    	if (auth == null) {
    		throw new ApiException(HttpStatus.FORBIDDEN, ErrorMessage.withCode("AUTH.BAD_USERNAME_OR_PASSWORD"));
    	}
    	return new LoginResponse("Authorization", "Basic " + Base64.getEncoder().encodeToString((login.getUsername() + ":" + login.getPassword()).getBytes()));
    }
    
    @RequestMapping(path = "/auth/signup", method = RequestMethod.POST)
    public LoginResponse signup(@RequestBody Signup signup, Locale locale) {
    	if (!signupEnabled) {
    		throw new ResponseStatusException(HttpStatus.NOT_FOUND);
    	}
    	//validate
    	if (signup == null) {
    		throw new ResponseStatusException(HttpStatus.BAD_REQUEST);
    	}
    	
    	List<ErrorMessage> errors = new ArrayList<>();
    	
    	//TODO validate email
    	if (!StringUtils.hasText(signup.getLogin())) {
    		errors.add(ErrorMessage.withFieldCode("login", "AUTH.BAD_USERNAME"));
    	}

		if (!StringUtils.hasText(signup.getEmail())) {
			errors.add(ErrorMessage.withFieldCode("email", "AUTH.BAD_EMAIL"));
		}

    	if (!PasswordUtils.validatePassword(signup.getPassword())) {
    		errors.add(ErrorMessage.withFieldCode("password", "AUTH.BAD_PASSWORD"));
    	}
    	
    	if (!Boolean.TRUE.equals(signup.getTerms())) {
    		errors.add(ErrorMessage.withFieldCode("terms", "AUTH.TERMS_REQUIRED"));
    	}
    	if (!Boolean.TRUE.equals(signup.getPrivacy())) {
    		errors.add(ErrorMessage.withFieldCode("privacy", "AUTH.PRIVACY_REQUIRED"));
    	}
    	
    	if (!errors.isEmpty()) {
    		throw new ApiException(HttpStatus.BAD_REQUEST, errors);
    	}
    	
    	try {
    		authService.register(signup.getLogin(), signup.getEmail(), signup.getPassword(), locale);
    	} catch (BadCredentialsException e) {
    		errors.add(ErrorMessage.withCode(e.getMessage()));
    	} catch (Exception e) {
    		errors.add(ErrorMessage.withMessage(e.getMessage()));
    		e.printStackTrace();
    	}

    	if (errors.isEmpty()) {
    		return new LoginResponse("Authorization", "Basic " + Base64.getEncoder().encodeToString((signup.getLogin() + ":" + signup.getPassword()).getBytes()));
    	}
    	
		throw new ApiException(HttpStatus.BAD_REQUEST, errors);
    }
    
    @RequestMapping(path = "/auth/resetRequest", method = RequestMethod.POST)
    public ResponseEntity<List<AuthError>> resetToken(@RequestBody ResetPassword request) {
    	if (!resetEnabled) {
    		return ResponseEntity.notFound().build();
    	}
    	Utils.sleep(250);
    	try {
    		authService.generateResetToken(request.getLogin());
    	} catch (BadCredentialsException e) {
    		throw new ResponseStatusException(HttpStatus.FORBIDDEN);
    	} catch (Exception e) {
    		throw new ResponseStatusException(HttpStatus.INTERNAL_SERVER_ERROR);
    	}
    	return ResponseEntity.accepted().build();
    }
    
    @PostMapping(path = "/auth/resetPassword")
    public ResponseEntity<List<AuthError>> reset(@RequestBody ResetPassword resetPassword) {
    	if (!resetEnabled) {
    		return ResponseEntity.notFound().build();
    	}
    	Utils.sleep(250);
    	String token = resetPassword.getToken();
    	if (!StringUtils.hasText(token)) {
    		return new ResponseEntity<>(Collections.singletonList(new AuthError(null, "AUTH.INVALID_TOKEN")), HttpStatus.BAD_REQUEST);
    	}
    	
    	List<ErrorMessage> errors = new ArrayList<>();
    	
    	try {
    		authService.resetPassword(resetPassword.getLogin(), resetPassword.getToken(), resetPassword.getPassword());
//    		if (errs == null || errs.isEmpty()) {
//    			Authentication auth = authService.authenticate(resetPassword.getLogin(), resetPassword.getPassword());
//    			SecurityContextHolder.getContext().setAuthentication(auth);
//    		} else {
//    			errors.addAll(errs);
//    		}
    	} catch (Exception e) {
    		e.printStackTrace();
    		errors.add(ErrorMessage.withMessage(e.getMessage()));
    	}
    	
    	if (errors.isEmpty()) {
    		//return no errors
    		return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    	}
    	
		throw new ApiException(HttpStatus.BAD_REQUEST, errors);
    }
    
    @GetMapping("/auth/user")
    public AppUser user(@AuthenticationPrincipal AppUser user) {
    	//TODO other params
    	return user;
    }

    @DeleteMapping("/auth/user")
    public ResponseEntity<Void> deleteUser() {
    	Authentication auth = SecurityContextHolder.getContext().getAuthentication();
    	
    	AppUser user = null;
    	if (auth instanceof AppUserAuthentication) {
    		user = ((AppUserAuthentication) auth).getUser();
    	}
    	if (user == null) {
    		return ResponseEntity.notFound().build();
    	}
    	authService.delete(user);
    	return ResponseEntity.accepted().build();
    }

    @PutMapping("/auth/password")
    public ResponseEntity<Void> changePassword(@AuthenticationPrincipal AppUser user, @RequestBody ChangePassword chp) {
    	if (user != null) {
    		authService.changePassword(user.getLogin(), chp.getCurrentPassword(), chp.getPassword());
    		return ResponseEntity.accepted().build();
    	}
    	
    	return ResponseEntity.notFound().build();
    }

}
